@@ -7,6 +7,7 @@
 #include <hyprutils/memory/WeakPtr.hpp>
 #include <wayland-client.h>
 #include <xf86drmMode.h>
+#include <optional>
 
 namespace Aquamarine {
     class CDRMBackend;
@@ -14,6 +15,7 @@ namespace Aquamarine {
     class CDRMOutput;
     struct SDRMConnector;
     class CDRMRenderer;
+    class CDRMDumbAllocator;
 
     typedef std::function<void(void)> FIdleCallback;
 
@@ -162,6 +164,7 @@ namespace Aquamarine {
             bool     ownModeID = false;
             uint32_t modeID    = 0;
             uint32_t gammaLut  = 0;
+            uint32_t ctm       = 0;
         } atomic;
 
         Hyprutils::Memory::CSharedPointer<SDRMPlane> primary;
@@ -175,6 +178,9 @@ namespace Aquamarine {
                 uint32_t vrr_enabled;
                 uint32_t gamma_lut;
                 uint32_t gamma_lut_size;
+                uint32_t ctm;
+                uint32_t degamma_lut;
+                uint32_t degamma_lut_size;
 
                 // atomic-modesetting only
 
@@ -182,7 +188,7 @@ namespace Aquamarine {
                 uint32_t mode_id;
                 uint32_t out_fence_ptr;
             };
-            uint32_t props[7] = {0};
+            uint32_t props[9] = {0};
         };
         UDRMCRTCProps props;
     };
@@ -194,11 +200,12 @@ namespace Aquamarine {
         virtual bool                                                      test();
         virtual Hyprutils::Memory::CSharedPointer<IBackendImplementation> getBackend();
         virtual bool                                                      setCursor(Hyprutils::Memory::CSharedPointer<IBuffer> buffer, const Hyprutils::Math::Vector2D& hotspot);
-        virtual void                                                      moveCursor(const Hyprutils::Math::Vector2D& coord, bool skipShedule = false);
+        virtual void                                                      moveCursor(const Hyprutils::Math::Vector2D& coord, bool skipSchedule = false);
         virtual void                                                      scheduleFrame(const scheduleFrameReason reason = AQ_SCHEDULE_UNKNOWN);
         virtual void                                                      setCursorVisible(bool visible);
         virtual Hyprutils::Math::Vector2D                                 cursorPlaneSize();
         virtual size_t                                                    getGammaSize();
+        virtual size_t                                                    getDeGammaSize();
         virtual std::vector<SDRMFormat>                                   getRenderFormats();
 
         int                                                               getConnectorID();
@@ -242,13 +249,21 @@ namespace Aquamarine {
         uint32_t                                  flags    = 0;
         bool                                      test     = false;
         drmModeModeInfo                           modeInfo;
+        std::optional<Hyprutils::Math::Mat3x3>    ctm;
+        std::optional<hdr_output_metadata>        hdrMetadata;
 
         struct {
-            uint32_t gammaLut = 0;
-            uint32_t fbDamage = 0;
-            uint32_t modeBlob = 0;
-            bool     blobbed  = false;
-            bool     gammad   = false;
+            uint32_t gammaLut   = 0;
+            uint32_t degammaLut = 0;
+            uint32_t fbDamage   = 0;
+            uint32_t modeBlob   = 0;
+            uint32_t ctmBlob    = 0;
+            uint32_t hdrBlob    = 0;
+            bool     blobbed    = false;
+            bool     gammad     = false;
+            bool     degammad   = false;
+            bool     ctmd       = false;
+            bool     hdrd       = false; // true if hdr blob needs updating or clearing
         } atomic;
 
         void calculateMode(Hyprutils::Memory::CSharedPointer<SDRMConnector> connector);
@@ -262,7 +277,7 @@ namespace Aquamarine {
         void                                           disconnect();
         Hyprutils::Memory::CSharedPointer<SDRMCRTC>    getCurrentCRTC(const drmModeConnector* connector);
         drmModeModeInfo*                               getCurrentMode();
-        void                                           parseEDID(std::vector<uint8_t> data);
+        IOutput::SParsedEDID                           parseEDID(std::vector<uint8_t> data);
         bool                                           commitState(SDRMConnectorCommitData& data);
         void                                           applyCommit(const SDRMConnectorCommitData& data);
         void                                           rollbackCommit(const SDRMConnectorCommitData& data);
@@ -308,58 +323,75 @@ namespace Aquamarine {
                 uint32_t vrr_capable;  // not guaranteed to exist
                 uint32_t subconnector; // not guaranteed to exist
                 uint32_t non_desktop;
-                uint32_t panel_orientation; // not guaranteed to exist
-                uint32_t content_type;      // not guaranteed to exist
-                uint32_t max_bpc;           // not guaranteed to exist
+                uint32_t panel_orientation;   // not guaranteed to exist
+                uint32_t content_type;        // not guaranteed to exist
+                uint32_t max_bpc;             // not guaranteed to exist
+                uint32_t Colorspace;          // not guaranteed to exist
+                uint32_t hdr_output_metadata; // not guaranteed to exist
 
                 // atomic-modesetting only
 
                 uint32_t crtc_id;
             };
-            uint32_t props[4] = {0};
+            uint32_t props[13] = {0};
         };
         UDRMConnectorProps props;
+
+        union UDRMConnectorColorspace {
+            struct {
+                uint32_t Default;
+                uint32_t BT2020_RGB;
+                uint32_t BT2020_YCC;
+            };
+            uint32_t props[3] = {0};
+        };
+        UDRMConnectorColorspace colorspace;
     };
 
     class IDRMImplementation {
       public:
+        virtual ~IDRMImplementation()                                                                                  = default;
         virtual bool commit(Hyprutils::Memory::CSharedPointer<SDRMConnector> connector, SDRMConnectorCommitData& data) = 0;
         virtual bool reset()                                                                                           = 0;
 
         // moving a cursor IIRC is almost instant on most hardware so we don't have to wait for a commit.
-        virtual bool moveCursor(Hyprutils::Memory::CSharedPointer<SDRMConnector> connector, bool skipShedule = false) = 0;
+        virtual bool moveCursor(Hyprutils::Memory::CSharedPointer<SDRMConnector> connector, bool skipSchedule = false) = 0;
     };
 
     class CDRMBackend : public IBackendImplementation {
       public:
         virtual ~CDRMBackend();
-        virtual eBackendType                                            type();
-        virtual bool                                                    start();
-        virtual std::vector<Hyprutils::Memory::CSharedPointer<SPollFD>> pollFDs();
-        virtual int                                                     drmFD();
-        virtual bool                                                    dispatchEvents();
-        virtual uint32_t                                                capabilities();
-        virtual bool                                                    setCursor(Hyprutils::Memory::CSharedPointer<IBuffer> buffer, const Hyprutils::Math::Vector2D& hotspot);
-        virtual void                                                    onReady();
-        virtual std::vector<SDRMFormat>                                 getRenderFormats();
-        virtual std::vector<SDRMFormat>                                 getCursorFormats();
-        virtual bool                                                    createOutput(const std::string& name = "");
-        virtual Hyprutils::Memory::CSharedPointer<IAllocator>           preferredAllocator();
-        virtual std::vector<SDRMFormat>                                 getRenderableFormats();
+        virtual eBackendType                                               type();
+        virtual bool                                                       start();
+        virtual std::vector<Hyprutils::Memory::CSharedPointer<SPollFD>>    pollFDs();
+        virtual int                                                        drmFD();
+        virtual bool                                                       dispatchEvents();
+        virtual uint32_t                                                   capabilities();
+        virtual bool                                                       setCursor(Hyprutils::Memory::CSharedPointer<IBuffer> buffer, const Hyprutils::Math::Vector2D& hotspot);
+        virtual void                                                       onReady();
+        virtual std::vector<SDRMFormat>                                    getRenderFormats();
+        virtual std::vector<SDRMFormat>                                    getCursorFormats();
+        virtual bool                                                       createOutput(const std::string& name = "");
+        virtual Hyprutils::Memory::CSharedPointer<IAllocator>              preferredAllocator();
+        virtual std::vector<SDRMFormat>                                    getRenderableFormats();
+        virtual std::vector<Hyprutils::Memory::CSharedPointer<IAllocator>> getAllocators();
 
-        Hyprutils::Memory::CWeakPointer<CDRMBackend>                    self;
+        Hyprutils::Memory::CWeakPointer<CDRMBackend>                       self;
 
-        void                                                            log(eBackendLogLevel, const std::string&);
-        bool                                                            sessionActive();
-        int                                                             getNonMasterFD();
+        void                                                               log(eBackendLogLevel, const std::string&);
+        bool                                                               sessionActive();
+        int                                                                getNonMasterFD();
 
-        std::vector<FIdleCallback>                                      idleCallbacks;
-        std::string                                                     gpuName;
+        std::vector<FIdleCallback>                                         idleCallbacks;
+        std::string                                                        gpuName;
 
       private:
         CDRMBackend(Hyprutils::Memory::CSharedPointer<CBackend> backend);
 
         static std::vector<Hyprutils::Memory::CSharedPointer<CDRMBackend>> attempt(Hyprutils::Memory::CSharedPointer<CBackend> backend);
+        static Hyprutils::Memory::CSharedPointer<CDRMBackend>              fromGpu(std::string path, Hyprutils::Memory::CSharedPointer<CBackend> backend,
+                                                                                   Hyprutils::Memory::CSharedPointer<CDRMBackend> primary);
+
         bool registerGPU(Hyprutils::Memory::CSharedPointer<CSessionDevice> gpu_, Hyprutils::Memory::CSharedPointer<CDRMBackend> primary_ = {});
         bool checkFeatures();
         bool initResources();
@@ -369,6 +401,7 @@ namespace Aquamarine {
         void scanConnectors();
         void scanLeases();
         void restoreAfterVT();
+        void recheckOutputs();
         void recheckCRTCs();
         void buildGlFormats(const std::vector<SGLFormat>& fmts);
 
@@ -389,6 +422,8 @@ namespace Aquamarine {
         std::vector<SDRMFormat>                                       formats;
         std::vector<SDRMFormat>                                       glFormats;
 
+        Hyprutils::Memory::CSharedPointer<CDRMDumbAllocator>          dumbAllocator;
+
         bool                                                          atomic = false;
 
         struct {
